/* Aftershock 3D rendering engine
 * Copyright (C) 1999 Stephen C. Taylor
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "util.h"
#include "pak.h"
#include "bsp.h"
#include "render.h"
#include "tex.h"
#include "lightmap.h"
#include "shader.h"
#include "mesh.h"
#include "skybox.h"
#include "entity.h"
#include "mapent.h"
#include "uicommon.h"
#include "glinc.h"
#include <stdio.h>
#include <string.h>

#ifdef WIN32
#include "glext.h"
#endif

#include "globalshared.h"
extern global_shared_t *g;

#define AZ_SCALE 0.1
#define EL_SCALE 0.1

static int ox, oy;
static double fpstime;
static int fpsframes;
static int invert_mouse = 0;

static void
find_start_pos(void)
{
    /* Find the first spawn point in the entities list */
    int i;
    const char *cname;
    
    g->r_eye_el = g->r_eye_az = 0.0;
    g->r_eyepos[0] = g->r_eyepos[1] = g->r_eyepos[2] = 0.0;
    for (i = 0; i < g->g_numentities; i++)
    {
	cname = entity_value(i, "classname");
	if (!strcmp(cname, "info_player_deathmatch"))
	{
	    g->r_eye_az = entity_float(i, "angle");
	    entity_vec3(i, "origin", g->r_eyepos);
	    break;
	}
    }
    vec_copy(g->r_eyepos, g->r_eyepos_sav)
}

void updateVelocity(float *vel, int mov, double etime, float accel, float max) {
    if (etime > 0.1) etime = 0.1; /* cap */

    *vel += mov * accel * etime;

    if (*vel > max)  *vel = max;
    if (*vel < -max) *vel = -max;

    /* slow down */
    if (*vel > 0.0) {
      *vel -= max/2 * etime;
      if (*vel < 0.0) *vel = 0.0;
    }
    else if (*vel < 0.0) {
      *vel += max/2 * etime;
      if (*vel > 0.0) *vel = 0.0;
    }
}


void
move_eye(double etime)
{
    vec3_t delta;
    vec3_t aheadPos, aheadDelta;
    int aheadCluster;
    float aheadD;

    if (g->r_setup_projection)
      updateVelocity(&(g->g_velocity), g->g_move, etime, 1000.0, 400.0);
    else
      updateVelocity(&(g->g_velocity), g->g_move, etime, 200.0, 200.0);

    if (g->g_velocity > 0)
      aheadD = g->g_velocity * etime + g->g_collision_buffer;
    else
      aheadD = g->g_velocity * etime - g->g_collision_buffer;

    vec_copy(g->r_eyepos, aheadPos);
    vec_copy(g->r_movedir, aheadDelta);
    vec_scale(aheadDelta, aheadD, aheadDelta);
    vec_add(aheadPos, aheadDelta, aheadPos);

    aheadCluster = find_cluster(aheadPos);
    if (g->g_collision && aheadCluster < 0) {
      g->g_velocity = -(g->g_velocity) * 0.6;
     return;
    }

    vec_copy(g->r_movedir, delta);
    vec_scale(delta, g->g_velocity * etime, delta);
    vec_add(g->r_eyepos, delta, g->r_eyepos);
}

static float
calc_fov(float fov_x, float width, float height)
{
    /* Borrowed from the Quake 1 source distribution */
    float   a;
    float   x;
    
    if (fov_x < 1 || fov_x > 179)
	Error("Bad fov: %f", fov_x);
    
    x = width/tan(fov_x/360.0*M_PI);
    a = atan(height/x);
    a = a*360.0/M_PI;
    
    return a;
}

/* reads the command line arguments */
void
ui_read_args(int argc, char **argv, char **bsp, char **pak)
{
	int i;
	int bsp_specified=FALSE;
	char *arg;

	for(i=1; i < argc; i++) {
		arg=argv[i];
		if (arg[0] == '-')
		{
		/* this is a switch argument */

		if(!strcmp(arg,"-f") || !strcmp(arg,"--fullscreen"))
			g->r_fullscreen=TRUE;
		else if(!strcmp(arg,"--notex"))
			g->r_notextures=TRUE;
		else if(!strcmp(arg,"--noitems"))
			g->r_drawitems=FALSE;
	} else {
		/* BSP or pak file specified */
		if(bsp_specified)
		{
			*pak=arg;
		} else {
			bsp_specified=TRUE;
			*bsp=arg;
		}
	  }
    }
}

void ui_init_gl(r_context_t *context)
{
    float fov_y;

    /* fprintf(stderr, " Initializing rendering context... \n"); */

    /* My GL driver (linux/TNT2) has problems with glEnableClientState(),
       but this seems to clear it up.  Go figure. */
    glFinish();
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    context->r_tex_units = 1; /* default - 1 texture unit */
#ifndef NO_MULTITEX
    glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &(context->r_tex_units));
#endif

    render_init(context);
    lightmap_bindobjs(context);
    tex_bindobjs(context);

    glDisable(GL_DITHER);
    glShadeModel(GL_SMOOTH);
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glColor3f(1.0, 1.0, 1.0);

    glEnable(GL_DEPTH_TEST);
    glCullFace(GL_FRONT);

    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    if (g->r_setup_projection) {
      float aspect = ((float) g->r_viewport_w)/((float) g->r_viewport_h);
      glViewport(0, 0, g->r_viewport_w, g->r_viewport_h);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      fov_y = calc_fov(g->r_eyefov, g->r_viewport_w, g->r_viewport_h);
      gluPerspective(fov_y, aspect, 5.0, 3000.0); /* was 10.0, 3000.0 */
      glMatrixMode(GL_MODELVIEW);
    }

    /* printf(" done\n\n"); */
}

void print_info()
{
  printf("\n");
  printf("    C A V E   Q U A K E   I I I   A R E N A \n");
  printf("    --------------------------------------- \n");
  printf("               by Paul Rajlich \n");
  printf("\n");
  printf("               paul@visbox.com \n");
  printf("         http://www.visbox.com/cq3a/ \n");
  printf("\n");
  printf(" Based on the Aftershock engine by Steve Taylor.\n");
  printf("   Quake3 Arena is a trademark of Id Software.\n");
  printf("\n");
}

void ui_init_bsp(char *bsp, char *pak)
{
    g->g_frametime = 0;
    fpstime = 0;
    fpsframes = 0;

    print_info();

    fprintf(stderr, " Loading bsp...");
    pak_openpak(pak);
    bsp_read(bsp);
    fprintf(stderr, " done\n");

    fprintf(stderr, " Creating meshes...");
    mesh_create_all();
    fprintf(stderr, " done\n");

    skybox_create();
 
    fprintf(stderr, " Loading entities...");
    mapent_loadall();
    render_init_mapents();
    fprintf(stderr, " done\n");

    fprintf(stderr, " Parsing shaders...");
    shader_readall();
    fprintf(stderr, " done\n");

    fprintf(stderr, " Loading lightmaps...");
    lightmap_loadall();
    fprintf(stderr, " done\n");

    fprintf(stderr, "\n");
    fprintf(stderr, " Loading textures\n\n");
    tex_loadall();
    fprintf(stderr, "\n");

    find_start_pos();
    vec_point(g->r_movedir, g->r_eye_az, g->r_eye_el);
}

void
ui_move(double time)
{
    static double lastTime = 0.0f;
    double etime = time - lastTime;

    /* cap */
    if (etime <= 0.0) etime = 0.0;
    if (etime > 0.1) etime = 0.1;

    /* move! */
    move_eye(etime);

    if (g->g_collision) {
      if (g->r_eyecluster >= 0) {
        vec_copy(g->r_eyepos, g->r_eyepos_sav)
      }
      else {
        vec_copy(g->r_eyepos_sav, g->r_eyepos)
        g->r_eyecluster = find_cluster(g->r_eyepos);
      }
    }

    /* FPS counter */
    fpsframes++;
    if (time - fpstime > 0.5) /* was 1.0 */
    {
        g->r_fps = fpsframes / (time - fpstime);
        fprintf(stderr,"  FPS: %.1f   sky: %d   collision ", g->r_fps, g->r_sky);
        if (g->g_collision)
          fprintf(stderr,"on   \r");
        else
          fprintf(stderr,"off  \r");
        fpstime = time;
        fpsframes = 0;
    }

    lastTime = time;
}

void
ui_display(r_context_t *c, double time)
{
    g->g_frametime = time;

    if (!(g->r_lockpvs))
      g->r_eyecluster = find_cluster(g->r_eyepos);
    
    render_scene(c, time);
    /* glFlush(); */
}

void
ui_mouse_down(int x, int y)
{
    ox = x; oy = y;
}

void
ui_mouse_motion(int x, int y)
{
    g->r_eye_az += -(x - ox) * AZ_SCALE;
    g->r_eye_el += -(y - oy) * EL_SCALE * (invert_mouse ? -1.0f : 1.0f);

    if (g->r_eye_el >  90.0) g->r_eye_el =  90.0;
    if (g->r_eye_el < -90.0) g->r_eye_el = -90.0;

    vec_point(g->r_movedir, g->r_eye_az, g->r_eye_el);
    
    ox = x; oy = y;    
}

void
ui_mouse_relmotion(int relx, int rely)
{
    g->r_eye_az += -relx * AZ_SCALE;
    g->r_eye_el += -rely * EL_SCALE * (invert_mouse ? -1.0f : 1.0f);

    if (g->r_eye_el >  90.0) g->r_eye_el =  90.0;
    if (g->r_eye_el < -90.0) g->r_eye_el = -90.0;

    vec_point(g->r_movedir, g->r_eye_az, g->r_eye_el);
}
    


void
ui_key_down(unsigned char key)
{
    switch (key)
    {
	case 'w':
	    g->g_move = 1;
	    break;
	case 's':
	    g->g_move = -1;
	    break;
	case 'b':
    	    g->r_sky++;
	    if (g->r_sky > 2)
		    g->r_sky = 0;
            if (g->r_sky == 1)
              glClearColor(0.65, 0.65, 1.0, 1.0); 
            if (g->r_sky == 2)
              glClearColor(0.0, 0.0, 0.0, 1.0);
	    break;
	case 'c':
	    if (g->g_collision)
	      g->g_collision = 0;
	    else	    
	      g->g_collision = 1;
	    break;	
        case 'p':
         printf("\rpos: %.1f %.1f %.1f az: %.1f el: %.1f dir: %.3f %.3f %.3f\n",
			    g->r_eyepos[0], g->r_eyepos[1], g->r_eyepos[2],
			    g->r_eye_az, g->r_eye_el,
			    g->r_movedir[0], g->r_movedir[1], g->r_movedir[2]);
	    break;
		/***
	case 'l':
	    g->r_lockpvs = !(g->r_lockpvs);
	    if (g->r_lockpvs) printf("cluster: %d\n", g->r_eyecluster);
	    break;
	case 'i':
	    invert_mouse = !invert_mouse;
	    printf("invert_mouse = %d\n", invert_mouse);
	    break;
		***/
	case 'q':
		exit(0);
		break;
    }
}

void
ui_key_up(unsigned char key)
{
    switch (key)
    {
	case 'w':
	case 's':
	    g->g_move = 0;
	    break;
    }
}

void
ui_finish(r_context_t *c)
{
    shader_freeall();
    mesh_free_all();
    skybox_free();
    mapent_freeall();
    lightmap_freeobjs(c);
    tex_freeobjs(c);
    render_finalize(c);
    pak_closepak();
    fprintf(stderr, "\n\n");
}
