/* vrJuggler interface
 * Copyright (C) 2001 Paul Rajlich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <GL/gl.h>

// vrJuggler stuff
#include <Kernel/vjKernel.h>
#include <Kernel/GL/vjGlApp.h>
#include <Kernel/GL/vjGlContextData.h>
#include <Kernel/vjDebug.h>
#include <Kernel/vjProjection.h>
#include <Input/InputManager/vjPosInterface.h>
#include <Input/InputManager/vjDigitalInterface.h>
#include <Input/InputManager/vjKeyboardInterface.h>
#include <Input/InputManager/vjAnalogInterface.h>
#include <Performance/vjTimeStamp.h>

// timing
#include <sys/time.h>
#include <unistd.h>

extern "C" {

#include "vec.h"
#include "uicommon.h"
#include "rendercontext.h"
#include "globalshared.h"

};

global_shared_t *g;

int stopRequested=0;

void printUsage();

// use regular memory (juggler uses threads)
void *gc_malloc(size_t size) { return malloc(size); }
void *rc_malloc(size_t size) { return malloc(size); }
/*
// timing
#ifdef WIN32
double gettime()
{
   return ((double)GetTickCount())/1000.0f;
}
#else
double gettime()
{
    struct timeval tv;
    gettimeofday(&tv, 0);
    return  ((double) tv.tv_sec + (double) tv.tv_usec / 1000000.0);
}
#endif
*/

// Wrapper class to hold a copy of a context specific data structure
class ContextData
{
public:
   ContextData() : r_context(NULL)
   {
      r_context = (r_context_t *) rc_malloc(sizeof(r_context_t));
      std::cout << "Allocated ContextData." << std::endl;
   }

   r_context_t* getData()
   { return r_context; }

private:
   r_context_t* r_context;
};



// Actual application class
// Just derives from vjGlApp base class and implments the interface methods
// needed to call the rendering engine.
class myApp : public vjGlApp {
private:
  // see rendercontext.h. Each rendering thread will have its own copy
  vjGlContextData<ContextData> contextData;

public:
  /*
  vjPosInterface mWand;
  vjDigitalInterface mButton0;
  vjDigitalInterface mButton1;
  vjDigitalInterface mButton2;
  */
  vjKeyboardInterface mKB;
  vjAnalogInterface   mTime;

#define NKEYS 6

  struct key
  {
    int vjc;
    char uic;
    bool on;
    void init (int vj,char ui,bool _on=false) { vjc=vj; uic=ui; on=_on; }
  };

  key keys[NKEYS];


  //double starttime, frametime, lastTime;
  double totaltime;
  char *bsp, *pak;

public:
  myApp(vjKernel *kern, char *bsp, char *pak) : vjGlApp(kern) {
    this->bsp = strdup(bsp);
    this->pak = strdup(pak);
    std::cerr << std::endl;
    std::cerr << "***************************************" << std::endl;
    std::cerr << "bsp: " << bsp << " pak: " << pak << std::endl;
    std::cerr << "***************************************" << std::endl;
    std::cerr << std::endl << std::flush;

    keys[0].init(VJKEY_UP,'w');
    keys[1].init(VJKEY_DOWN,'s');
    keys[2].init(VJKEY_C,'c');
    keys[3].init(VJKEY_B,'b');
    keys[4].init(VJMBUTTON1,'w');
    keys[5].init(VJMBUTTON3,'s');
  }

   // Initialize global data structures
  virtual void init() {

    //mWand.init("VJWand");
    //mButton0.init("VJButton0");
    //mButton1.init("VJButton1");
    //mButton2.init("VJButton2");
    mKB.init("Keyboard");
    mTime.init("Time");
    std::cout << "input ok\n"<<std::flush;

    // setup global data structure
    g = (global_shared_t *) gc_malloc(sizeof(global_shared_t));
    init_global_shared(g);

    // Some global parameters to tweak
    //g->r_gamma = 1.7;
    //g->r_notextures = 1;       // Turn off textures for better performance

    // load the map
    ui_init_bsp(bsp, pak);
    printf("\n\n");
    totaltime=0;
    g->r_setup_projection = 0;
  }

  // Initialize context data
  virtual void contextInit() {
    r_context_t* the_context_data = contextData->getData();
    vjASSERT(the_context_data != NULL && "context data pointer should be non-NULL");

    ui_init_gl(the_context_data);
  }

  // Execute the actually drawing
  virtual void draw() {
    r_context_t* the_context_data = contextData->getData();
    vjASSERT(the_context_data != NULL);

    // Set some state that is expected
    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    glCullFace(GL_FRONT);
    glShadeModel(GL_SMOOTH);

    ui_display(the_context_data, totaltime);
    glDisable(GL_BLEND);

    // reset some state to make simulator mode look "better"
    glCullFace(GL_BACK);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glShadeModel(GL_FLAT);
  }

  // Updates that occur immedately after device updates but before rendering
  // - Update timing
  // - Update navigation
  virtual void preFrame() {

    if (mKB->keyPressed(VJKEY_CTRL) && mKB->keyPressed(VJKEY_ESC))
      stopRequested = 1;

    for (int i=0;i<NKEYS;i++)
      {
	bool on=mKB->keyPressed(keys[i].vjc);
	if (on && !keys[i].on)
	  ui_key_down(keys[i].uic);
	else if (!on && keys[i].on)
	  ui_key_up(keys[i].uic);
	keys[i].on=on;
      }

    totaltime+=mTime->getData();

    ui_mouse_relmotion(mKB->keyPressed(VJMOUSE_POSX)-mKB->keyPressed(VJMOUSE_NEGX)
	              ,mKB->keyPressed(VJMOUSE_POSY)-mKB->keyPressed(VJMOUSE_NEGY));

    ui_move(totaltime);


  }
};


int
main(int argc, char **argv)
{
  // defaults
  char* pak = "pak0.pk3";
  char* bsp = "maps/q3tourney2.bsp";

  // USAGE:
  // command -c configfile1 -c configfile2 -- otherOptions
  // NOTE: The config files all have to be first then the "--" specification
  if(argc <= 1)
  {
    printUsage();
    return 1;
  }
  vjKernel* kernel = vjKernel::instance();
  kernel->parseArg(&argc,&argv);
/*
  // Get the Juggler commands first
  int c;
  while ( (c = getopt(argc, argv, "hc:")) != -1)
  {
      switch (c) {
      case 'c':
           kernel->loadConfigFile( std::string(optarg) );
           break;
      case 'h':
            printUsage();
            exit(1);
            break;
      case '?':
           break;
      }
  }

  // optind-1 because it will ignore elt 0 (normally app name)
  ui_read_args( (argc-optind+1), &(argv[optind-1]), &bsp, &pak );
*/
  int i;
  for (i=1;i<argc;i++)
  {
    if (!strcmp(argv[i],"-")) break;
    kernel->loadConfigFile( std::string(argv[i]) );
  }
  if (i==argc)
    ui_read_args(1,argv,&bsp,&pak);
  else
    ui_read_args(argc-i,argv+i,&bsp,&pak);

  std::cout <<"bsp="<<bsp<<" pak="<<pak<<std::endl<<std::flush;

  myApp *app = new myApp(kernel, bsp, pak);

  kernel->start();
  kernel->setApplication(app);

  while (!stopRequested) {
    usleep(100000);
  }
}

void printUsage()
{
   std::cout << "Usage: cq3a <juggler-config-file> <more-juggler-config> - <cq3a config params>" << std::endl;
}

